# Some of this test code is specific to 2.7
import pathlib
import sys
sys.path.insert(0, "..")
import unittest
from pytwobit import TwoBit

class TwoBitLocalTest(unittest.TestCase):

    def setUp(self):
        # Edit URL as appropriate
        self.filename = str((pathlib.Path(__file__).parent / "test.2bit").resolve())


    def test_init(self):
        t = TwoBit(self.filename)
        self.assertTrue(isinstance(t, TwoBit))
        self.assertEqual(t.sequenceCount, 1)
        self.assertIsNotNone(t.index)

    def test_getSequenceRecord(self):
        t = TwoBit(self.filename)
        sequence_record = t.sequence_record('chr1')
        self.assertEqual(sequence_record["dnaSize"], 159)

    def test_getSequence(self):
        twobit = TwoBit(self.filename)

        #Non-masked no "N" region
        expectedSeq = "CTATCTA"
        seq =  twobit.fetch("chr1", 50, 57)
        self.assertEqual(expectedSeq, seq)

        #"N" region
        expectedSeq = "NNNNN"
        seq =  twobit.fetch("chr1", 42, 47)
        self.assertEqual(expectedSeq, seq)

        # Mixed region
        expectedSeq = "NNNACT"
        seq =  twobit.fetch("chr1", 44, 50)
        self.assertEqual(expectedSeq, seq)

        # partially masked region
        expectedSeq = "tagcatcctcctacctcacNNacCNctTGGACNCcCaGGGatttcNNNcc"
        seq =  twobit.fetch("chr1", 100, 150)
        self.assertEqual(expectedSeq, seq)


class TwoBitRemoteTest(unittest.TestCase):

    def test_twobit_sequence(self):
        url = "https://igv.org/genomes/data/hg38/hg38.2bit"
        twobit = TwoBit(url)

        #Non-masked no "N" region  chr1:11,830-11,869
        expectedSeq = "GATTGCCAGCACCGGGTATCATTCACCATTTTTCTTTTCG"
        seq =  twobit.fetch("chr1", 11829, 11869)
        self.assertEqual(expectedSeq, seq)

        #"N" region  chr1:86-124
        expectedSeq = "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN"
        seq =  twobit.fetch("chr1", 85, 124)
        self.assertEqual(expectedSeq, seq)

        # partially masked region chr1:120,565,295-120,565,335
        expectedSeq = "TATGAACTTTTGTTCGTTGGTgctcagtcctagaccctttt"
        seq =  twobit.fetch("chr1", 120565294, 120565335)
        self.assertEqual(expectedSeq, seq)

        # Unrecongized sequence name
        seq =  twobit.fetch("noSuchSequence", 0)
        self.assertIsNone(seq)




if __name__ == '__main__':
    unittest.main()





class DataView:

    def __init__(self, bytes, byte_order):

        self.bytes = bytes
        self.position = 0
        self.byte_order = byte_order


    def getByte(self):
        val = self.bytes[self.position]
        self.position = self.position + 1
        return val

    def getString(self, length):
        val = self.bytes[self.position:self.position+length].decode('utf8')
        self.position = self.position + length
        return val

    def uint32(self):

        val =  int.from_bytes(self.bytes[self.position:self.position+4], self.byte_order)
        self.position = self.position + 4
        return val


    def available(self):

        return len(self.bytes) - self.position


import requests

class RemoteFile:

    """
    A "file like" object for random access to a remote binary file.   Used in place of a local file
    handle in 'array.fromfile' calls.  Includes a simple buffer to redunce the number of http requests
    for small bits of data.
    """
    def __init__(self, url):
        self.url = url
        self.offset = 0

    def seek(self, offet):
        self.offset = offet

    def tell(self):
        return self.offset

    def close(self):
        self.offset = 0

    def read(self, n):

        if n == 0:
            return b''

        end = self.offset + n - 1
        headers = {
            "Content-Type":"application/octet-stream",
            "Range":f"bytes={self.offset}-{end}"}
        response = requests.get(self.url, headers = headers)
        bytes = response.content
        return bytes

    def size(self):

        response = requests.head(self.url)
        sz = response.headers.get('content-length')
        return int(sz) if sz is not None else None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass



from bisect import bisect_left
from urllib.parse import urlparse
import math
from .remote_file import RemoteFile
from .data_view import DataView

twoBit = ['T', 'C', 'A', 'G']
byteTo4Bases = []
for i in range(0, 256):
    byteTo4Bases.append(
        twoBit[(i >> 6) & 3] +
        twoBit[(i >> 4) & 3] +
        twoBit[(i >> 2) & 3] +
        twoBit[i & 3])

maskedByteTo4Bases = []
for i in range(0, 256):
    maskedByteTo4Bases.append(byteTo4Bases[i].lower())


class TwoBit:

    def __init__(self, path):

        self.path = path
        self.init()

    def open_file_handle(self):
        if urlparse(self.path).scheme in ['http', 'https']:
            return RemoteFile(self.path)
        else:
            return open(self.path, 'rb')

    def check_magic(self, dataview):
        magic = dataview.uint32()
        if magic != 0x1A412743:
            self.byte_order = 'big'
            dataview.byte_order = 'big'
            dataview.position = 0
            magic = dataview.uint32()
            if magic != 0x1A412743:
                raise Exception("Bad magic number")

    def init(self):

        with self.open_file_handle() as file_handle:

            self.meta_index = {}

            self.byte_order = 'little'  # Assumption, will be checked

            dataview = self._dataview(file_handle, 0, 1024)

            self.check_magic(dataview) # Checks magic, side effect - advances dataview cursor

            self.version = dataview.uint32()
            self.sequenceCount = dataview.uint32()
            self.reserved = dataview.uint32()

            # Read index
            index = {}
            estNameLength = 20
            ptr = 0
            for i in range(0, self.sequenceCount):

                if dataview.available() < 1:
                    ptr = ptr + dataview.position
                    size = (self.sequenceCount - i) * estNameLength
                    dataview = self._dataview(file_handle, ptr, size)

                strlen = dataview.getByte()
                if dataview.available() < strlen + 5:
                    ptr = ptr + dataview.position
                    size = (self.sequenceCount - i) * estNameLength
                    dataview = self._dataview(file_handle, ptr, size)

                name = dataview.getString(strlen)
                offset = dataview.uint32()
                index[name] = offset

            self.index = index

    def sequence_record(self, seqName):

        with self.open_file_handle() as file_handle:

            if seqName not in self.meta_index:

                if seqName not in self.index:
                    return None

                offset = self.index[seqName]
                dataview = self._dataview(file_handle, offset, 8)
                dnaSize = dataview.uint32()
                nBlockCount = dataview.uint32()

                offset = offset + 8

                # Read "N" blocks
                size = nBlockCount * (4 + 4) + 4
                dataview = self._dataview(file_handle, offset, size)

                nBlockStarts = []
                for i in range(0, nBlockCount):
                    nBlockStarts.append(dataview.uint32())

                nBlockSizes = []
                for i in range(0, nBlockCount):
                    nBlockSizes.append(dataview.uint32())

                # Read "mask" blocks
                maskBlockCount = dataview.uint32()
                offset += size

                size = maskBlockCount * (4 + 4) + 4
                dataview = self._dataview(file_handle, offset, size)

                maskBlockStarts = []
                for i in range(0, maskBlockCount):
                    maskBlockStarts.append(dataview.uint32())

                maskBlockSizes = []
                for i in range(0, maskBlockCount):
                    maskBlockSizes.append(dataview.uint32())

                # Transform "N" and "mask" block data into something more useful
                nBlocks = []
                for i in range(0, nBlockCount):
                    nBlocks.append(Block(nBlockStarts[i],nBlockSizes[i]))

                maskBlocks = []
                for i in range(0, maskBlockCount):
                    maskBlocks.append(Block(maskBlockStarts[i],maskBlockSizes[i]))

                reserved = dataview.uint32()
                if reserved != 0:
                    raise 'Bad 2-bit file'

                offset += size
                meta = {
                    "dnaSize": dnaSize,
                    "nBlocks": nBlocks,
                    "maskBlocks": maskBlocks,
                    "packedPos": offset,
                    "bpLength": dnaSize
                }

                self.meta_index[seqName] = meta

            return self.meta_index[seqName]

    def fetch(self, seqName, regionStart=None, regionEnd=None):

        with self.open_file_handle() as file_handle:

            record = self.sequence_record(seqName)
            if record is None:
                return None

            if regionStart is None:
                regionStart = 0
            elif regionStart < 0:
                raise 'regionStart cannot be less than 0'


            # end defaults to the end of the sequence
            if regionEnd is None or regionEnd > record["dnaSize"]:
                regionEnd = record["dnaSize"]

            # Get the "N" and "mask" blocks
            nBlocks = getOverlappingBlocks(regionStart, regionEnd, record["nBlocks"])
            maskBlocks = getOverlappingBlocks(regionStart, regionEnd, record["maskBlocks"])

            baseBytesOffset = math.floor(regionStart / 4)
            start = record["packedPos"] + baseBytesOffset
            size = math.floor(regionEnd / 4) - baseBytesOffset + 1

            file_handle.seek(start)
            baseBytes = file_handle.read(size)

            sequenceBases = []

            genomicPosition = regionStart
            while genomicPosition < regionEnd:

                # check if we are currently masked
                # trim masks to the left of genomic position
                n = next((i for i, block in enumerate(maskBlocks) if block.end > genomicPosition), None)
                if n is not None and n > 0:
                    maskBlocks = maskBlocks[n:]

                baseIsMasked = len(maskBlocks) > 0 and maskBlocks[0].start <= genomicPosition and maskBlocks[0].end > genomicPosition

                # process the N block if we have one.  Masked "N" ("n")  is not supported
                if len(nBlocks) > 0 and genomicPosition >= nBlocks[0].start and genomicPosition < nBlocks[0].end:
                    currentNBlock = nBlocks[0]
                    nBlocks = nBlocks[1:]
                    n_count = min(currentNBlock.end, regionEnd) - genomicPosition
                    sequenceBases.extend(['N'] * n_count)
                    genomicPosition += n_count
                    genomicPosition = genomicPosition - 1
                else:
                    bytePosition = math.floor(genomicPosition / 4) - baseBytesOffset
                    subPosition = genomicPosition % 4
                    byte = baseBytes[bytePosition]
                    if baseIsMasked:
                        sequenceBases.append(maskedByteTo4Bases[byte][subPosition])
                    else:
                        sequenceBases.append(byteTo4Bases[byte][subPosition])

                genomicPosition = genomicPosition + 1

            seqstring = ''.join(sequenceBases)
            return seqstring


    def _dataview(self, file_handle, offset, size):
        file_handle.seek(offset)
        bytes = file_handle.read(size)
        return DataView(bytes, self.byte_order)

class Block:
    def __init__(self, start, size):
        self.start = start
        self.end = start + size


def getOverlappingBlocks(start, end, blocks):

    overlappingBlocks = []

    for block in blocks:
        if block.start > end:
            break
        elif block.end < start:
            continue
        else:
            overlappingBlocks.append(block)

    return overlappingBlocks




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyTwobit',
    version='0.2.0',
    description='A fast reader for local or remote UCSC twobit sequence files.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    author='Jim Robinson',
    url='https://github.com/jrobinso/pyTwoBit',
    packages=['pytwobit'],
    package_data={'pytwobit': ['tests/foo.2bit']},
    install_requires=[
        'requests',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
)
